@@ -30,8 +30,8 @@ pub enum Reconfig {
 impl std::fmt::Debug for Reconfig {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         match self {
-            Reconfig::Join(a) => write!(f, "J{:?}", a),
-            Reconfig::Leave(a) => write!(f, "L{:?}", a),
+            Reconfig::Join(a) => write!(f, "J{}", a),
+            Reconfig::Leave(a) => write!(f, "L{}", a),
         }
     }
 }
@@ -118,7 +118,7 @@ pub struct SignedVote {
 
 impl Debug for SignedVote {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        write!(f, "{:?}@{:?}", self.vote, self.voter)
+        write!(f, "{:?}@{}", self.vote, self.voter)
     }
 }
 
@@ -404,7 +404,7 @@ impl State {
         Ok(vec![])
     }
 
-    fn sign_vote(&self, vote: Vote) -> Result<SignedVote, Error> {
+    pub fn sign_vote(&self, vote: Vote) -> Result<SignedVote, Error> {
         Ok(SignedVote {
             voter: self.public_key(),
             sig: self.secret_key.sign(&vote.to_bytes()?),
@@ -495,6 +495,7 @@ impl State {
 
     pub fn validate_signed_vote(&self, signed_vote: &SignedVote) -> Result<(), Error> {
         signed_vote.validate_signature()?;
+        self.validate_vote(&signed_vote.vote)?;
 
         let members = self.members(self.gen)?;
         if !members.contains(&signed_vote.voter) {
@@ -509,12 +510,7 @@ impl State {
             Err(Error::ExistingVoteIncompatibleWithNewVote {
                 existing_vote: self.votes[&signed_vote.voter].clone(),
             })
-        } else if self.pending_gen == self.gen {
-            // We are starting a vote for the next generation
-            self.validate_vote(&signed_vote.vote)
         } else {
-            // This is a vote for this generation
-
             // Ensure that nobody is trying to change their reconfig's.
             let reconfigs: BTreeSet<(PublicKey, Reconfig)> = self
                 .votes
@@ -527,7 +523,7 @@ impl State {
             if voters.len() != reconfigs.len() {
                 Err(Error::VoterChangedMind { reconfigs })
             } else {
-                self.validate_vote(&signed_vote.vote)
+                Ok(())
             }
         }
     }
